/*
 * Segment 2 (KEx),  SWE 2.WebFS Web Feature Service
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package org.geotools.data.vrz3;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URI;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.geotools.data.AbstractDataStore;
import org.geotools.data.FeatureReader;
import org.geotools.feature.FeatureType;
import org.geotools.xml.gml.GMLSchema;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavParameters;
import de.bsvrz.dav.daf.main.CommunicationError;
import de.bsvrz.dav.daf.main.ConnectionException;
import de.bsvrz.dav.daf.main.InconsistentLoginException;
import de.bsvrz.dav.daf.main.MissingParameterException;
import de.bsvrz.kex.wfs.wfs.ExportObjektDefinition;
import de.bsvrz.kex.wfs.wfs.WFSException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * DataStore Implementation f&uuml;r die DaV-Daten.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: VRZ3DataStore.java 8819 2008-05-09 14:58:18Z gieseler $
 * 
 */
public class VRZ3DataStore extends AbstractDataStore {

	/**
	 * DebugLogger fr Debug-Ausgaben.
	 */
	private static final Debug DEBUG = Debug.getLogger();

	/**
	 * DataStore-Parameter.
	 */
	private HashMap<String, Object> params = null;

	/**
	 * Feature-Map.
	 */
	private final HashMap<String, VRZ3Feature> featureMap = new HashMap<String, VRZ3Feature>();

	/** Namespace. */
	private URI namespace;

	/** Parameter der Datenverteiler-Verbindung. */
	private String davParam;

	/** DaV-Benutzer. */
	private String davBenutzer;

	/** DaV-Authentifizierung. */
	private String davAuthentifizierung;

	/** Dateiname der Exportdatei. */
	private String exportclassname;

	/** Datenverteiler-Verbindung. */
	private ClientDavConnection dav;

	/**
	 * ERzeugt einen VRZ3-DataStore.
	 * 
	 * @param params
	 *            Parameter
	 * @throws WFSException
	 *             bei Ausnahmen
	 * 
	 */
	public VRZ3DataStore(HashMap<String, Object> params) throws WFSException {
		super(false); // NOT writable
		// System.err.println("Classpath: " +
		// System.getProperty("java.class.path"));
		DEBUG.info("Konstruktor");
		this.params = (params != null) ? params : new HashMap<String, Object>();
		initialisiere(params);
	}

	/**
	 * Gibt einen FeatureReader f&uuml;r ein Feature zur&uuml;ck.
	 * 
	 * @param typeName
	 *            FeatureTyp Name
	 * 
	 * @return FeatureReader
	 */
	@Override
	protected FeatureReader getFeatureReader(String typeName) {
		if (featureMap.containsKey(typeName)) {
			return featureMap.get(typeName).getFeatureReader();
		}

		return null;
	}

	// /**
	// * Gibt einen FeatureWriter f&uuml;r ein Feature zur&uuml;ck.
	// *
	// * @param typeName
	// * FeatureTyp Name
	// *
	// * @return FeatureWriter
	// * @throws IOException
	// * bei Ausnahmen
	// */
	// @Override
	// protected FeatureWriter getFeatureWriter(String typeName)
	// throws IOException {
	// return null;
	// }

	/**
	 * Gibt den Wert eines Parameters zur&uuml;ck.
	 * 
	 * @param name
	 *            Name des Parameters
	 * @param defaultwert
	 *            Standardwert
	 * @param required
	 *            Parameter obligatorisch ja/nein
	 * @param paramsmap
	 *            Parameter
	 * @return Wert des Parameters
	 * 
	 * 
	 * @throws IOException
	 *             if required parameter is missing
	 */
	private Object getParam(String name, Object defaultwert, boolean required,
			Map<String, Object> paramsmap) throws IOException {
		Object result;

		try {
			result = params.get(name);
		} catch (Exception e) {
			result = null;
		}

		if (result == null) {
			if (required) {
				throw new IOException("VRZ3DataStore: parameter " + name
						+ " wird bentigt");
			}

			result = defaultwert;
		}

		return result;
	}

	/**
	 * Gibt das Schema eines Feature-Typs zur&uuml;ck.
	 * 
	 * @param typeName
	 *            Feature-Typ Name
	 * 
	 * @return FeatureType
	 * @throws IOException
	 *             bei Ausnahmen
	 */
	@Override
	public FeatureType getSchema(String typeName) throws IOException {
		if (featureMap.containsKey(typeName)) {
			return featureMap.get(typeName).getSchema();
		}

		return null;
	}

	/**
	 * Gibt die Liste der Feature-Typ Namen zur&uuml;ck.
	 * 
	 * @return Liste der Feature-Typ Namen
	 * 
	 * @throws IOException
	 *             bei Ausnahmen
	 */
	@Override
	public String[] getTypeNames() throws IOException {
		// scanFiles(filePath); // re-scans path just in case some file was
		// added

		String[] names = new String[featureMap.size()];
		int index = 0;

		for (Iterator<String> i = featureMap.keySet().iterator(); i.hasNext();) {
			names[index++] = i.next();
		}

		return names;
	}

	/**
	 * Initialisiert den DataStore.
	 * 
	 * @param parameter
	 *            Parameter
	 * @throws WFSException
	 *             bei Ausnahmen
	 */
	private void initialisiere(HashMap<String, Object> parameter)
			throws WFSException {
		// ArgumentList argList = new ArgumentList(new String[] { _dav_param });
		ClientDavParameters cdp;
		try {
			parseParameter(parameter);
			cdp = new ClientDavParameters();
			String[] p = davParam.split(":");
			cdp.setDavCommunicationAddress(p[0]);
			cdp.setDavCommunicationSubAddress(Integer.parseInt(p[1]));
			// cdp.setUserName(userName)
			// cdp.setUserPassword(userPassword)

			dav = new ClientDavConnection(cdp);
			dav.connect();
			dav.login(davBenutzer, davAuthentifizierung);
		} catch (MissingParameterException e) {
			throw new WFSException(e.getMessage());
		} catch (CommunicationError e) {
			throw new WFSException(e.getMessage());
		} catch (ConnectionException e) {
			throw new WFSException(e.getMessage());
		} catch (InconsistentLoginException e) {
			throw new WFSException(e.getMessage());
		} catch (Exception e) {
			throw new WFSException(e.getMessage());
		}

		initialisiereFeatures();
	}

	/**
	 * Initialisiert die Features des DataStores.
	 */
	private void initialisiereFeatures() {
		Class<?> exportobjekteclass;
		String exportobjekteclassname = exportclassname;

		// _debug.info("Die Netzreferenzen werden geprft."); //$NON-NLS-1$

		try {
			exportobjekteclass = Class.forName(exportobjekteclassname);
			Method getmethode = exportobjekteclass.getMethod(
					"getExportObjekte", new Class[] {}); //$NON-NLS-1$
			Object[][] exportobjekte = (Object[][]) getmethode.invoke(null,
					new Object[] {});
			Map<String, VRZ3Feature> fmap = new HashMap<String, VRZ3Feature>();

			// Featureliste aufbauen
			for (int i = 0; i < exportobjekte.length; ++i) {
				try {
					DEBUG.info("Erzeuge Feature: "
							+ exportobjekte[i][ExportObjektDefinition.TYP]);
					// mehrere Attributgruppen pro Typ sind erlaubt
					VRZ3Feature ff = fmap
							.get(exportobjekte[i][ExportObjektDefinition.TYP]);
					if (ff != null) {
						ff.hinzufuegenAttributgruppe(exportobjekte[i]);
					} else {
						fmap
								.put(
										(String) exportobjekte[i][ExportObjektDefinition.TYP],
										new VRZ3Feature(dav, exportobjekte[i],
												namespace));
					}
				} catch (Exception e) {
					// _debug.error(e.getMessage());
					e.printStackTrace();
				}
			}

			// Feature erzeugen
			for (VRZ3Feature f : fmap.values()) {
				try {
					f.buildFeature();
					featureMap.put(f.getSchema().getTypeName(), f);
				} catch (Exception e) {
					DEBUG.error(e.getMessage());
					e.printStackTrace();
				}
			}
		} catch (ClassNotFoundException e) {
			DEBUG
					.error("Die Klasse der zu exportierenden Objekte konnte nicht geladen werden: " //$NON-NLS-1$
							+ exportobjekteclassname);
			e.printStackTrace();
			throw new RuntimeException(e);
		} catch (NoSuchMethodException e) {
			DEBUG.error("Methode 'getExportObjekte()' in der Klasse '" //$NON-NLS-1$
					+ exportobjekteclassname + "' existiert nicht."); //$NON-NLS-1$
			e.printStackTrace();
			throw new RuntimeException(e);

		} catch (IllegalAccessException e) {
			DEBUG
					.error("Fehler beim Zugriff auf die Methode 'getExportObjekte()' in der Klasse '" //$NON-NLS-1$
							+ exportobjekteclassname + "'."); //$NON-NLS-1$ 
			e.printStackTrace();
			throw new RuntimeException(e);

		} catch (InvocationTargetException e) {
			// _debug.error(e.getMessage());
			e.printStackTrace();
			throw new RuntimeException(e);
		}
	}

	/**
	 * Initialisiert die Variablen aus den Parametern.
	 * 
	 * @param parameter
	 *            Parameter
	 * @throws IOException
	 *             bei Ausnahmen
	 */
	private void parseParameter(Map<String, Object> parameter)
			throws IOException {
		// if (parameter == null) {
		// parameter = new HashMap<String, Object>();
		// }

		namespace = (URI) getParam(VRZ3DataStoreFactory.PARAM_NAMESPACE.key,
				GMLSchema.NAMESPACE, true, parameter);

		davParam = (String) getParam(VRZ3DataStoreFactory.PARAM_DAV.key,
				VRZ3DataStoreFactory.PARAM_DAV.sample, true, parameter);

		davBenutzer = (String) getParam(
				VRZ3DataStoreFactory.PARAM_DAV_NUTZER.key,
				VRZ3DataStoreFactory.PARAM_DAV_NUTZER.sample, true, parameter);

		davAuthentifizierung = (String) getParam(
				VRZ3DataStoreFactory.PARAM_DAV_AUTHENTIFIZIERUNG.key,
				VRZ3DataStoreFactory.PARAM_DAV_AUTHENTIFIZIERUNG.sample, true,
				parameter);

		exportclassname = (String) getParam(
				VRZ3DataStoreFactory.PARAM_EXPORTCLASS.key,
				VRZ3DataStoreFactory.PARAM_EXPORTCLASS.sample, false, parameter);
	}
}
